@@ -116,7 +116,9 @@ json_object *LUKS2_get_keyslot_jobj(struct luks2_hdr *hdr, int keyslot)
 	if (!json_object_object_get_ex(hdr->jobj, "keyslots", &jobj1))
 		return NULL;
 
-	json_object_object_get_ex(jobj1, keyslot_name, &jobj2);
+	if (!json_object_object_get_ex(jobj1, keyslot_name, &jobj2))
+		return NULL;
+
 	return jobj2;
 }
 
@@ -167,24 +169,50 @@ json_object *LUKS2_get_digest_jobj(struct luks2_hdr *hdr, int digest)
 	return jobj2;
 }
 
+static json_object *json_get_segments_jobj(json_object *hdr_jobj)
+{
+	json_object *jobj_segments;
+
+	if (!hdr_jobj || !json_object_object_get_ex(hdr_jobj, "segments", &jobj_segments))
+		return NULL;
+
+	return jobj_segments;
+}
+
 json_object *LUKS2_get_segment_jobj(struct luks2_hdr *hdr, int segment)
 {
-	json_object *jobj1, *jobj2;
-	char segment_name[16];
-
-	if (!hdr || segment < 0)
+	if (!hdr)
 		return NULL;
 
-	if (snprintf(segment_name, sizeof(segment_name), "%u", segment) < 1)
-		return NULL;
+	if (segment == CRYPT_DEFAULT_SEGMENT)
+		segment = LUKS2_get_default_segment(hdr);
 
-	if (!json_object_object_get_ex(hdr->jobj, "segments", &jobj1))
-		return NULL;
+	return json_segments_get_segment(json_get_segments_jobj(hdr->jobj), segment);
+}
 
-	if (!json_object_object_get_ex(jobj1, segment_name, &jobj2))
-		return NULL;
+json_object *LUKS2_get_segments_jobj(struct luks2_hdr *hdr)
+{
+	return hdr ? json_get_segments_jobj(hdr->jobj) : NULL;
+}
 
-	return jobj2;
+int LUKS2_segments_count(struct luks2_hdr *hdr)
+{
+	if (!hdr)
+		return -EINVAL;
+
+	return json_segments_count(LUKS2_get_segments_jobj(hdr));
+}
+
+int LUKS2_get_default_segment(struct luks2_hdr *hdr)
+{
+	int s = LUKS2_get_segment_id_by_flag(hdr, "backup-final");
+	if (s >= 0)
+		return s;
+
+	if (LUKS2_segments_count(hdr) == 1)
+		return 0;
+
+	return -EINVAL;
 }
 
 /*
@@ -216,7 +244,8 @@ static json_bool json_str_to_uint64(json_object *jobj, uint64_t *value)
 uint64_t json_object_get_uint64(json_object *jobj)
 {
 	uint64_t r;
-	return json_str_to_uint64(jobj, &r) ? r : 0;
+	json_str_to_uint64(jobj, &r);
+	return r;
 }
 
 json_object *json_object_new_uint64(uint64_t value)
@@ -264,35 +293,7 @@ json_object *json_contains(struct crypt_device *cd, json_object *jobj, const cha
 	return sobj;
 }
 
-/* use only on already validated 'segments' object */
-static uint64_t get_first_data_offset(json_object *jobj_segs, const char *type)
-{
-	json_object *jobj_offset, *jobj_type;
-	uint64_t tmp, min = UINT64_MAX;
-
-	json_object_object_foreach(jobj_segs, key, val) {
-		UNUSED(key);
-
-		if (type) {
-			json_object_object_get_ex(val, "type", &jobj_type);
-			if (strcmp(type, json_object_get_string(jobj_type)))
-				continue;
-		}
-
-		json_object_object_get_ex(val, "offset", &jobj_offset);
-		tmp = json_object_get_uint64(jobj_offset);
-
-		if (!tmp)
-			return tmp;
-
-		if (tmp < min)
-			min = tmp;
-	}
-
-	return min;
-}
-
-static json_bool validate_json_uint32(json_object *jobj)
+json_bool validate_json_uint32(json_object *jobj)
 {
 	int64_t tmp;
 
@@ -402,7 +403,7 @@ static json_bool validate_intervals(struct crypt_device *cd,
 	return TRUE;
 }
 
-int LUKS2_keyslot_validate(struct crypt_device *cd, json_object *hdr_jobj, json_object *hdr_keyslot, const char *key)
+static int LUKS2_keyslot_validate(struct crypt_device *cd, json_object *hdr_jobj, json_object *hdr_keyslot, const char *key)
 {
 	json_object *jobj_key_size;
 
@@ -562,18 +563,48 @@ static int hdr_validate_crypt_segment(struct crypt_device *cd,
 	return !segment_has_digest(key, jobj_digests);
 }
 
+static bool validate_segment_intervals(struct crypt_device *cd,
+				    int length, const struct interval *ix)
+{
+	int j, i = 0;
+
+	while (i < length) {
+		if (ix[i].length == UINT64_MAX && (i != (length - 1))) {
+			log_dbg(cd, "Only last regular segment is allowed to have 'dynamic' size.");
+			return false;
+		}
+
+		for (j = 0; j < length; j++) {
+			if (i == j)
+				continue;
+			if ((ix[i].offset >= ix[j].offset) && (ix[j].length == UINT64_MAX || (ix[i].offset < (ix[j].offset + ix[j].length)))) {
+				log_dbg(cd, "Overlapping segments [%" PRIu64 ",%" PRIu64 "]%s and [%" PRIu64 ",%" PRIu64 "]%s.",
+					ix[i].offset, ix[i].offset + ix[i].length, ix[i].length == UINT64_MAX ? "(dynamic)" : "",
+					ix[j].offset, ix[j].offset + ix[j].length, ix[j].length == UINT64_MAX ? "(dynamic)" : "");
+				return false;
+			}
+		}
+
+		i++;
+	}
+
+	return true;
+}
+
 static int hdr_validate_segments(struct crypt_device *cd, json_object *hdr_jobj)
 {
-	json_object *jobj, *jobj_digests, *jobj_offset, *jobj_size, *jobj_type, *jobj_flags;
-	int i;
+	json_object *jobj_segments, *jobj_digests, *jobj_offset, *jobj_size, *jobj_type, *jobj_flags, *jobj;
+	struct interval *intervals;
 	uint64_t offset, size;
+	int i, r, count, first_backup = -1;
 
-	if (!json_object_object_get_ex(hdr_jobj, "segments", &jobj)) {
+	if (!json_object_object_get_ex(hdr_jobj, "segments", &jobj_segments)) {
 		log_dbg(cd, "Missing segments section.");
 		return 1;
 	}
 
-	if (json_object_object_length(jobj) < 1) {
+	count = json_object_object_length(jobj_segments);
+	if (count < 1) {
 		log_dbg(cd, "Empty segments section.");
 		return 1;
 	}
@@ -582,7 +613,7 @@ static int hdr_validate_segments(struct crypt_device *cd, json_object *hdr_jobj)
 	if (!json_object_object_get_ex(hdr_jobj, "digests", &jobj_digests))
 		return 1;
 
-	json_object_object_foreach(jobj, key, val) {
+	json_object_object_foreach(jobj_segments, key, val) {
 		if (!numbered(cd, "Segment", key))
 			return 1;
 
@@ -623,12 +654,61 @@ static int hdr_validate_segments(struct crypt_device *cd, json_object *hdr_jobj)
 					return 1;
 		}
 
+		i = atoi(key);
+		if (json_segment_is_backup(val)) {
+			if (first_backup < 0 || i < first_backup)
+				first_backup = i;
+		} else {
+			if ((first_backup >= 0) && i >= first_backup) {
+				log_dbg(cd, "Regular segment at %d is behind backup segment at %d", i, first_backup);
+				return 1;
+			}
+		}
+
 		/* crypt */
 		if (!strcmp(json_object_get_string(jobj_type), "crypt") &&
 		    hdr_validate_crypt_segment(cd, val, key, jobj_digests, offset, size))
 			return 1;
 	}
 
+	if (first_backup == 0) {
+		log_dbg(cd, "No regular segment.");
+		return 1;
+	}
+
+	if (first_backup < 0)
+		first_backup = count;
+
+	intervals = malloc(first_backup * sizeof(*intervals));
+	if (!intervals) {
+		log_dbg(cd, "Not enough memory.");
+		return 1;
+	}
+
+	for (i = 0; i < first_backup; i++) {
+		jobj = json_segments_get_segment(jobj_segments, i);
+		if (!jobj) {
+			log_dbg(cd, "Gap at key %d in segments object.", i);
+			free(intervals);
+			return 1;
+		}
+		intervals[i].offset = json_segment_get_offset(jobj, 0);
+		intervals[i].length = json_segment_get_size(jobj, 0) ?: UINT64_MAX;
+	}
+
+	r = !validate_segment_intervals(cd, first_backup, intervals);
+	free(intervals);
+
+	if (r)
+		return 1;
+
+	for (; i < count; i++) {
+		if (!json_segments_get_segment(jobj_segments, i)) {
+			log_dbg(cd, "Gap at key %d in segments object.", i);
+			return 1;
+		}
+	}
+
 	return 0;
 }
 
@@ -681,6 +761,7 @@ static int hdr_validate_areas(struct crypt_device *cd, json_object *hdr_jobj)
 	json_object_object_foreach(jobj_keyslots, key, val) {
 
 		if (!(jobj_area = json_contains(cd, val, key, "Keyslot", "area", json_type_object)) ||
+		    !json_contains(cd, jobj_area, key, "Keyslot area", "type", json_type_string) ||
 		    !(jobj_offset = json_contains(cd, jobj_area, key, "Keyslot", "offset", json_type_string)) ||
 		    !(jobj_length = json_contains(cd, jobj_area, key, "Keyslot", "size", json_type_string)) ||
 		    !numbered(cd, "offset", json_object_get_string(jobj_offset)) ||
@@ -783,8 +864,7 @@ static int hdr_validate_config(struct crypt_device *cd, json_object *hdr_jobj)
 	 * validate keyslots_size fits in between (2 * metadata_size) and first
 	 * segment_offset (except detached header)
 	 */
-	json_object_object_get_ex(hdr_jobj, "segments", &jobj);
-	segment_offset = get_first_data_offset(jobj, "crypt");
+	segment_offset = json_segments_get_minimal_offset(json_get_segments_jobj(hdr_jobj), 0);
 	if (segment_offset &&
 	    (segment_offset < keyslots_size ||
 	     (segment_offset - keyslots_size) < (2 * metadata_size))) {
@@ -873,8 +953,9 @@ int LUKS2_hdr_read(struct crypt_device *cd, struct luks2_hdr *hdr, int repair)
 		/* unlikely: auto-recovery is required and failed due to read lock being held */
 		device_read_unlock(cd, crypt_metadata_device(cd));
 
+		/* Do not use LUKS2_device_write lock. Recovery. */
 		r = device_write_lock(cd, crypt_metadata_device(cd));
-		if (r) {
+		if (r < 0) {
 			log_err(cd, _("Failed to acquire write lock on device %s."),
 				device_path(crypt_metadata_device(cd)));
 			return r;
@@ -889,6 +970,18 @@ int LUKS2_hdr_read(struct crypt_device *cd, struct luks2_hdr *hdr, int repair)
 	return r;
 }
 
+int LUKS2_hdr_write_force(struct crypt_device *cd, struct luks2_hdr *hdr)
+{
+	/* NOTE: is called before LUKS2 validation routines */
+	/* erase unused digests (no assigned keyslot or segment) */
+	LUKS2_digests_erase_unused(cd, hdr);
+
+	if (LUKS2_hdr_validate(cd, hdr->jobj, hdr->hdr_size - LUKS2_HDR_BIN_LEN))
+		return -EINVAL;
+
+	return LUKS2_disk_hdr_write(cd, hdr, crypt_metadata_device(cd), false);
+}
+
 int LUKS2_hdr_write(struct crypt_device *cd, struct luks2_hdr *hdr)
 {
 	/* NOTE: is called before LUKS2 validation routines */
@@ -898,7 +991,7 @@ int LUKS2_hdr_write(struct crypt_device *cd, struct luks2_hdr *hdr)
 	if (LUKS2_hdr_validate(cd, hdr->jobj, hdr->hdr_size - LUKS2_HDR_BIN_LEN))
 		return -EINVAL;
 
-	return LUKS2_disk_hdr_write(cd, hdr, crypt_metadata_device(cd));
+	return LUKS2_disk_hdr_write(cd, hdr, crypt_metadata_device(cd), true);
 }
 
 int LUKS2_hdr_uuid(struct crypt_device *cd, struct luks2_hdr *hdr, const char *uuid)
@@ -962,9 +1055,9 @@ int LUKS2_hdr_backup(struct crypt_device *cd, struct luks2_hdr *hdr,
 		     const char *backup_file)
 {
 	struct device *device = crypt_metadata_device(cd);
-	int r = 0, devfd = -1;
+	int fd, devfd, r = 0;
 	ssize_t hdr_size;
-	ssize_t buffer_size;
+	ssize_t ret, buffer_size;
 	char *buffer = NULL;
 
 	hdr_size = LUKS2_hdr_and_areas_size(hdr->jobj);
@@ -993,19 +1086,17 @@ int LUKS2_hdr_backup(struct crypt_device *cd, struct luks2_hdr *hdr,
 		return devfd == -1 ? -EINVAL : devfd;
 	}
 
-	if (read_blockwise(devfd, device_block_size(cd, device),
-			   device_alignment(device), buffer, hdr_size) < hdr_size) {
-		close(devfd);
+	if (read_lseek_blockwise(devfd, device_block_size(cd, device),
+			   device_alignment(device), buffer, hdr_size, 0) < hdr_size) {
 		device_read_unlock(cd, device);
 		crypt_safe_free(buffer);
 		return -EIO;
 	}
-	close(devfd);
 
 	device_read_unlock(cd, device);
 
-	devfd = open(backup_file, O_CREAT|O_EXCL|O_WRONLY, S_IRUSR);
-	if (devfd == -1) {
+	fd = open(backup_file, O_CREAT|O_EXCL|O_WRONLY, S_IRUSR);
+	if (fd == -1) {
 		if (errno == EEXIST)
 			log_err(cd, _("Requested header backup file %s already exists."), backup_file);
 		else
@@ -1013,13 +1104,14 @@ int LUKS2_hdr_backup(struct crypt_device *cd, struct luks2_hdr *hdr,
 		crypt_safe_free(buffer);
 		return -EINVAL;
 	}
-	if (write_buffer(devfd, buffer, buffer_size) < buffer_size) {
+	ret = write_buffer(fd, buffer, buffer_size);
+	close(fd);
+	if (ret < buffer_size) {
 		log_err(cd, _("Cannot write header backup file %s."), backup_file);
 		r = -EIO;
 	} else
 		r = 0;
 
-	close(devfd);
 	crypt_safe_free(buffer);
 	return r;
 }
@@ -1034,12 +1126,17 @@ static int reqs_reencrypt(uint32_t reqs)
 	return reqs & CRYPT_REQUIREMENT_OFFLINE_REENCRYPT;
 }
 
+static int reqs_reencrypt_online(uint32_t reqs)
+{
+	return reqs & CRYPT_REQUIREMENT_ONLINE_REENCRYPT;
+}
+
 int LUKS2_hdr_restore(struct crypt_device *cd, struct luks2_hdr *hdr,
 		     const char *backup_file)
 {
 	struct device *backup_device, *device = crypt_metadata_device(cd);
-	int r, devfd = -1, diff_uuid = 0;
-	ssize_t buffer_size = 0;
+	int r, fd, devfd = -1, diff_uuid = 0;
+	ssize_t ret, buffer_size = 0;
 	char *buffer = NULL, msg[1024];
 	struct luks2_hdr hdr_file;
 	struct luks2_hdr tmp_hdr = {};
@@ -1049,7 +1146,6 @@ int LUKS2_hdr_restore(struct crypt_device *cd, struct luks2_hdr *hdr,
 	if (r < 0)
 		return r;
 
-	/* FIXME: why lock backup device ? */
 	r = device_read_lock(cd, backup_device);
 	if (r) {
 		log_err(cd, _("Failed to acquire read lock on device %s."),
@@ -1068,7 +1164,7 @@ int LUKS2_hdr_restore(struct crypt_device *cd, struct luks2_hdr *hdr,
 	}
 
 	/* do not allow header restore from backup with unmet requirements */
-	if (LUKS2_unmet_requirements(cd, &hdr_file, 0, 1)) {
+	if (LUKS2_unmet_requirements(cd, &hdr_file, CRYPT_REQUIREMENT_ONLINE_REENCRYPT, 1)) {
 		log_err(cd, _("Forbidden LUKS2 requirements detected in backup %s."),
 			backup_file);
 		r = -ETXTBSY;
@@ -1082,20 +1178,20 @@ int LUKS2_hdr_restore(struct crypt_device *cd, struct luks2_hdr *hdr,
 		goto out;
 	}
 
-	devfd = open(backup_file, O_RDONLY);
-	if (devfd == -1) {
+	fd = open(backup_file, O_RDONLY);
+	if (fd == -1) {
 		log_err(cd, _("Cannot open header backup file %s."), backup_file);
 		r = -EINVAL;
 		goto out;
 	}
 
-	if (read_buffer(devfd, buffer, buffer_size) < buffer_size) {
+	ret = read_buffer(fd, buffer, buffer_size);
+	close(fd);
+	if (ret < buffer_size) {
 		log_err(cd, _("Cannot read header backup file %s."), backup_file);
 		r = -EIO;
 		goto out;
 	}
-	close(devfd);
-	devfd = -1;
 
 	r = LUKS2_hdr_read(cd, &tmp_hdr, 0);
 	if (r == 0) {
@@ -1143,9 +1239,9 @@ int LUKS2_hdr_restore(struct crypt_device *cd, struct luks2_hdr *hdr,
 
 	log_dbg(cd, "Storing backup of header (%zu bytes) to device %s.", buffer_size, device_path(device));
 
-	/* TODO: perform header restore on bdev in stand-alone routine? */
+	/* Do not use LUKS2_device_write lock for checking sequence id on restore */
 	r = device_write_lock(cd, device);
-	if (r) {
+	if (r < 0) {
 		log_err(cd, _("Failed to acquire write lock on device %s."),
 			device_path(device));
 		goto out;
@@ -1163,15 +1259,13 @@ int LUKS2_hdr_restore(struct crypt_device *cd, struct luks2_hdr *hdr,
 		goto out;
 	}
 
-	if (write_blockwise(devfd, device_block_size(cd, device),
-			    device_alignment(device), buffer, buffer_size) < buffer_size)
+	if (write_lseek_blockwise(devfd, device_block_size(cd, device),
+			    device_alignment(device), buffer, buffer_size, 0) < buffer_size)
 		r = -EIO;
 	else
 		r = 0;
 
 	device_write_unlock(cd, device);
-	/* end of TODO */
-
 out:
 	LUKS2_hdr_free(cd, hdr);
 	LUKS2_hdr_free(cd, &hdr_file);
@@ -1180,11 +1274,7 @@ out:
 	crypt_memzero(&tmp_hdr, sizeof(tmp_hdr));
 	crypt_safe_free(buffer);
 
-	if (devfd >= 0) {
-		device_sync(cd, device, devfd);
-		close(devfd);
-	}
-
+	device_sync(cd, device);
 
 	return r;
 }
@@ -1279,6 +1369,7 @@ static const struct  {
 	const char *description;
 } requirements_flags[] = {
 	{ CRYPT_REQUIREMENT_OFFLINE_REENCRYPT, "offline-reencrypt" },
+	{ CRYPT_REQUIREMENT_ONLINE_REENCRYPT, "online-reencrypt" },
 	{ 0, NULL }
 };
 
@@ -1334,7 +1425,7 @@ int LUKS2_config_get_requirements(struct crypt_device *cd, struct luks2_hdr *hdr
 	return 0;
 }
 
-int LUKS2_config_set_requirements(struct crypt_device *cd, struct luks2_hdr *hdr, uint32_t reqs)
+int LUKS2_config_set_requirements(struct crypt_device *cd, struct luks2_hdr *hdr, uint32_t reqs, bool commit)
 {
 	json_object *jobj_config, *jobj_requirements, *jobj_mandatory, *jobj;
 	int i, r = -EINVAL;
@@ -1390,7 +1481,7 @@ int LUKS2_config_set_requirements(struct crypt_device *cd, struct luks2_hdr *hdr
 	if (!json_object_object_length(jobj_requirements))
 		json_object_object_del(jobj_config, "requirements");
 
-	return LUKS2_hdr_write(cd, hdr);
+	return commit ? LUKS2_hdr_write(cd, hdr) : 0;
 err:
 	json_object_put(jobj_mandatory);
 	return r;
@@ -1472,7 +1563,7 @@ static void hdr_dump_keyslots(struct crypt_device *cd, json_object *hdr_jobj)
 		json_object_object_get_ex(val, "type", &jobj2);
 		tmps = json_object_get_string(jobj2);
 
-		r = LUKS2_keyslot_for_segment(crypt_get_hdr(cd, CRYPT_LUKS2), j, CRYPT_DEFAULT_SEGMENT);
+		r = LUKS2_keyslot_for_segment(crypt_get_hdr(cd, CRYPT_LUKS2), j, CRYPT_ONE_SEGMENT);
 		log_std(cd, "  %s: %s%s\n", slot, tmps, r == -ENOENT ? " (unbound)" : "");
 
 		if (json_object_object_get_ex(val, "key_size", &jobj2))
@@ -1629,36 +1720,97 @@ int LUKS2_hdr_dump(struct crypt_device *cd, struct luks2_hdr *hdr)
 	return 0;
 }
 
+int LUKS2_get_data_size(struct luks2_hdr *hdr, uint64_t *size, bool *dynamic)
+{
+	int sector_size;
+	json_object *jobj_segments, *jobj_size;
+	uint64_t tmp = 0;
+
+	if (!size || !json_object_object_get_ex(hdr->jobj, "segments", &jobj_segments))
+		return -EINVAL;
+
+	json_object_object_foreach(jobj_segments, key, val) {
+		UNUSED(key);
+		if (json_segment_is_backup(val))
+			continue;
+
+		json_object_object_get_ex(val, "size", &jobj_size);
+		if (!strcmp(json_object_get_string(jobj_size), "dynamic")) {
+			sector_size = json_segment_get_sector_size(val);
+			/* last dynamic segment must have at least one sector in size */
+			if (tmp)
+				*size = tmp + (sector_size > 0 ? sector_size : SECTOR_SIZE);
+			else
+				*size = 0;
+			if (dynamic)
+				*dynamic = true;
+			return 0;
+		}
+
+		tmp += json_object_get_uint64(jobj_size);
+	}
+
+	/* impossible, real device size must not be zero */
+	if (!tmp)
+		return -EINVAL;
+
+	*size = tmp;
+	if (dynamic)
+		*dynamic = false;
+	return 0;
+}
+
 uint64_t LUKS2_get_data_offset(struct luks2_hdr *hdr)
 {
-	json_object *jobj1;
+	crypt_reencrypt_info ri;
+	json_object *jobj;
 
-	if (!json_object_object_get_ex(hdr->jobj, "segments", &jobj1))
-		return 0;
+	ri = LUKS2_reenc_status(hdr);
+	if (ri == CRYPT_REENCRYPT_CLEAN || ri == CRYPT_REENCRYPT_CRASH) {
+		jobj = LUKS2_get_segment_by_flag(hdr, "backup-final");
+		if (jobj)
+			return json_segment_get_offset(jobj, 1);
+	}
 
-	return get_first_data_offset(jobj1, "crypt") / SECTOR_SIZE;
+	return json_segments_get_minimal_offset(LUKS2_get_segments_jobj(hdr), 1);
 }
 
 const char *LUKS2_get_cipher(struct luks2_hdr *hdr, int segment)
 {
-	json_object *jobj1, *jobj2, *jobj3;
-	char buf[16];
+	json_object *jobj_segment;
 
-	if (segment < 0 || snprintf(buf, sizeof(buf), "%u", segment) < 1)
+	if (!hdr)
 		return NULL;
 
-	if (!json_object_object_get_ex(hdr->jobj, "segments", &jobj1))
-		return NULL;
+	if (segment == CRYPT_DEFAULT_SEGMENT)
+		segment = LUKS2_get_default_segment(hdr);
 
-	if (!json_object_object_get_ex(jobj1, buf, &jobj2))
+	jobj_segment = json_segments_get_segment(json_get_segments_jobj(hdr->jobj), segment);
+	if (!jobj_segment)
 		return NULL;
 
-	if (json_object_object_get_ex(jobj2, "encryption", &jobj3))
-		return json_object_get_string(jobj3);
-
 	/* FIXME: default encryption (for other segment types) must be string here. */
-	return "null";
+	return json_segment_get_cipher(jobj_segment) ?: "null";
+}
 
+crypt_reencrypt_info LUKS2_reenc_status(struct luks2_hdr *hdr)
+{
+	uint32_t reqs;
+
+	/*
+	 * Any unknown requirement or offline reencryption should abort
+	 * anything related to online-reencryption handling
+	 */
+	if (LUKS2_config_get_requirements(NULL, hdr, &reqs))
+		return CRYPT_REENCRYPT_INVALID;
+
+	if (!reqs_reencrypt_online(reqs))
+		return CRYPT_REENCRYPT_NONE;
+
+	if (json_segments_segment_in_reencrypt(LUKS2_get_segments_jobj(hdr)) < 0)
+		return CRYPT_REENCRYPT_CLEAN;
+
+	return CRYPT_REENCRYPT_CRASH;
 }
 
 const char *LUKS2_get_keyslot_cipher(struct luks2_hdr *hdr, int keyslot, size_t *key_size)
@@ -1689,28 +1841,23 @@ const char *LUKS2_get_keyslot_cipher(struct luks2_hdr *hdr, int keyslot, size_t
 
 const char *LUKS2_get_integrity(struct luks2_hdr *hdr, int segment)
 {
-	json_object *jobj1, *jobj2, *jobj3, *jobj4;
-	char buf[16];
+	json_object *jobj1, *jobj2, *jobj3;
 
-	if (segment < 0 || snprintf(buf, sizeof(buf), "%u", segment) < 1)
+	jobj1 = LUKS2_get_segment_jobj(hdr, segment);
+	if (!jobj1)
 		return NULL;
 
-	if (!json_object_object_get_ex(hdr->jobj, "segments", &jobj1))
+	if (!json_object_object_get_ex(jobj1, "integrity", &jobj2))
 		return NULL;
 
-	if (!json_object_object_get_ex(jobj1, buf, &jobj2))
+	if (!json_object_object_get_ex(jobj2, "type", &jobj3))
 		return NULL;
 
-	if (!json_object_object_get_ex(jobj2, "integrity", &jobj3))
-		return NULL;
-
-	if (!json_object_object_get_ex(jobj3, "type", &jobj4))
-		return NULL;
-
-	return json_object_get_string(jobj4);
+	return json_object_get_string(jobj3);
 }
 
 /* FIXME: this only ensures that once we have journal encryption, it is not ignored. */
+/* implement segment count and type restrictions (crypt and only single crypt) */
 static int LUKS2_integrity_compatible(struct luks2_hdr *hdr)
 {
 	json_object *jobj1, *jobj2, *jobj3, *jobj4;
@@ -1719,7 +1866,7 @@ static int LUKS2_integrity_compatible(struct luks2_hdr *hdr)
 	if (!json_object_object_get_ex(hdr->jobj, "segments", &jobj1))
 		return 0;
 
-	if (!json_object_object_get_ex(jobj1, CRYPT_DEFAULT_SEGMENT_STR, &jobj2))
+	if (!(jobj2 = LUKS2_get_segment_jobj(hdr, CRYPT_DEFAULT_SEGMENT)))
 		return 0;
 
 	if (!json_object_object_get_ex(jobj2, "integrity", &jobj3))
@@ -1770,6 +1917,9 @@ int LUKS2_get_volume_key_size(struct luks2_hdr *hdr, int segment)
 	json_object *jobj_digests, *jobj_digest_segments, *jobj_digest_keyslots, *jobj1;
 	char buf[16];
 
+	if (segment == CRYPT_DEFAULT_SEGMENT)
+		segment = LUKS2_get_default_segment(hdr);
+
 	if (snprintf(buf, sizeof(buf), "%u", segment) < 1)
 		return -1;
 
@@ -1795,17 +1945,182 @@ int LUKS2_get_volume_key_size(struct luks2_hdr *hdr, int segment)
 
 int LUKS2_get_sector_size(struct luks2_hdr *hdr)
 {
-	json_object *jobj1, *jobj_segment;
+	json_object *jobj_segment;
 
 	jobj_segment = LUKS2_get_segment_jobj(hdr, CRYPT_DEFAULT_SEGMENT);
 	if (!jobj_segment)
 		return SECTOR_SIZE;
 
-	json_object_object_get_ex(jobj_segment, "sector_size", &jobj1);
-	if (!jobj1)
-		return SECTOR_SIZE;
+	return json_segment_get_sector_size(jobj_segment) ?: SECTOR_SIZE;
+}
 
-	return json_object_get_int(jobj1);
+int LUKS2_assembly_multisegment_dmd(struct crypt_device *cd,
+	struct luks2_hdr *hdr,
+	struct volume_key *vks,
+	json_object *jobj_segments,
+	struct crypt_dm_active_device *dmd)
+{
+	struct volume_key *vk;
+	json_object *jobj;
+	enum devcheck device_check;
+	int r;
+	unsigned s = 0;
+	uint64_t data_offset, segment_size, segment_offset, segment_start = 0;
+	struct dm_target *t = &dmd->segment;
+
+	if (dmd->flags & CRYPT_ACTIVATE_SHARED)
+		device_check = DEV_OK;
+	else
+		device_check = DEV_EXCL;
+
+	data_offset = LUKS2_reencrypt_data_offset(hdr, true);
+
+	r = device_block_adjust(cd, crypt_data_device(cd), device_check,
+			                                data_offset, &dmd->size, &dmd->flags);
+	if (r)
+		return r;
+
+	r = dm_targets_allocate(&dmd->segment, json_segments_count(jobj_segments));
+	if (r)
+		goto err;
+
+	r = -EINVAL;
+
+	while (t) {
+		jobj = json_segments_get_segment(jobj_segments, s);
+		if (!jobj) {
+			log_dbg(cd, "Internal error. Segment %u is null.", s);
+			r = -EINVAL;
+			goto err;
+		}
+
+		segment_offset = json_segment_get_offset(jobj, 1);
+		segment_size = json_segment_get_size(jobj, 1);
+		/* 'dynamic' length allowed in last segment only */
+		if (!segment_size && !t->next)
+			segment_size = dmd->size - segment_start;
+		if (!segment_size) {
+			log_dbg(cd, "Internal error. Wrong segment size %u", s);
+			r = -EINVAL;
+			goto err;
+		}
+
+		if (!strcmp(json_segment_type(jobj), "crypt")) {
+			vk = crypt_volume_key_by_id(vks, LUKS2_digest_by_segment(hdr, s));
+			if (!vk) {
+				log_err(cd, _("Missing key for dm-crypt segment %u"), s);
+				r = -EINVAL;
+				goto err;
+			}
+
+			r = dm_crypt_target_set(t, segment_start, segment_size,
+					crypt_data_device(cd), vk,
+					json_segment_get_cipher(jobj),
+					json_segment_get_iv_offset(jobj),
+					segment_offset, "none", 0,
+					json_segment_get_sector_size(jobj));
+			if (r) {
+				log_err(cd, _("Failed to set dm-crypt segment."));
+				goto err;
+			}
+		} else if (!strcmp(json_segment_type(jobj), "linear")) {
+			r = dm_linear_target_set(t, segment_start, segment_size, crypt_data_device(cd), segment_offset);
+			if (r) {
+				log_err(cd, _("Failed to set dm-linear segment."));
+				goto err;
+			}
+		} else {
+			r = -EINVAL;
+			goto err;
+		}
+
+		segment_start += segment_size;
+		t = t->next;
+		s++;
+	}
+
+	return r;
+err:
+	dm_targets_free(cd, dmd);
+	return r;
+}
+
+/* FIXME: This shares almost all code with activate_multi_custom */
+static int _reload_custom_multi(struct crypt_device *cd,
+	const char *name,
+	struct volume_key *vks,
+	json_object *jobj_segments,
+	uint64_t device_size,
+	uint32_t flags)
+{
+	int r;
+	struct luks2_hdr *hdr = crypt_get_hdr(cd, CRYPT_LUKS2);
+	struct crypt_dm_active_device dmd =  {
+		.uuid   = crypt_get_uuid(cd),
+		.size = device_size >> SECTOR_SHIFT
+	};
+
+	/* do not allow activation when particular requirements detected */
+	if ((r = LUKS2_unmet_requirements(cd, hdr, CRYPT_REQUIREMENT_ONLINE_REENCRYPT, 0)))
+		return r;
+
+	/* Add persistent activation flags */
+	if (!(flags & CRYPT_ACTIVATE_IGNORE_PERSISTENT))
+		LUKS2_config_get_flags(cd, hdr, &dmd.flags);
+
+	dmd.flags |= (flags | CRYPT_ACTIVATE_SHARED);
+
+	r = LUKS2_assembly_multisegment_dmd(cd, hdr, vks, jobj_segments, &dmd);
+	if (!r)
+		r = dm_reload_device(cd, name, &dmd, 0, 0);
+
+	dm_targets_free(cd, &dmd);
+	return r;
+}
+
+int LUKS2_reload(struct crypt_device *cd,
+	const char *name,
+	struct volume_key *vks,
+	uint64_t device_size,
+	uint32_t flags)
+{
+	if (crypt_get_integrity_tag_size(cd))
+		return -ENOTSUP;
+
+	return _reload_custom_multi(cd, name, vks,
+			LUKS2_get_segments_jobj(crypt_get_hdr(cd, CRYPT_LUKS2)), device_size, flags);
+}
+
+int LUKS2_activate_multi(struct crypt_device *cd,
+	const char *name,
+	struct volume_key *vks,
+	uint64_t device_size,
+	uint32_t flags)
+{
+	struct luks2_hdr *hdr = crypt_get_hdr(cd, CRYPT_LUKS2);
+	json_object *jobj_segments = LUKS2_get_segments_jobj(hdr);
+	int r;
+	struct crypt_dm_active_device dmd = {
+		.size	= device_size,
+		.uuid   = crypt_get_uuid(cd)
+	};
+
+	/* do not allow activation when particular requirements detected */
+	if ((r = LUKS2_unmet_requirements(cd, hdr, CRYPT_REQUIREMENT_ONLINE_REENCRYPT, 0)))
+		return r;
+
+	/* Add persistent activation flags */
+	if (!(flags & CRYPT_ACTIVATE_IGNORE_PERSISTENT))
+		LUKS2_config_get_flags(cd, hdr, &dmd.flags);
+
+	dmd.flags |= flags;
+
+	r = LUKS2_assembly_multisegment_dmd(cd, hdr, vks, jobj_segments, &dmd);
+	if (!r)
+		r = dm_create_device(cd, name, CRYPT_LUKS2, &dmd);
+
+	dm_targets_free(cd, &dmd);
+	return r;
 }
 
 int LUKS2_activate(struct crypt_device *cd,
@@ -1816,7 +2131,7 @@ int LUKS2_activate(struct crypt_device *cd,
 	int r;
 	struct luks2_hdr *hdr = crypt_get_hdr(cd, CRYPT_LUKS2);
 	struct crypt_dm_active_device dmdi = {}, dmd = {
-		.uuid   = crypt_get_uuid(cd),
+		.uuid   = crypt_get_uuid(cd)
 	};
 
 	/* do not allow activation when particular requirements detected */
@@ -1838,7 +2153,7 @@ int LUKS2_activate(struct crypt_device *cd,
 
 	if (crypt_get_integrity_tag_size(cd)) {
 		if (!LUKS2_integrity_compatible(hdr)) {
-			log_err(cd, "Unsupported device integrity configuration.");
+			log_err(cd, _("Unsupported device integrity configuration."));
 			return -EINVAL;
 		}
 
@@ -1846,6 +2161,8 @@ int LUKS2_activate(struct crypt_device *cd,
 		if (r)
 			return r;
 
+		dmdi.flags |= CRYPT_ACTIVATE_PRIVATE;
+		dmdi.uuid = dmd.uuid;
 		dmd.segment.u.crypt.offset = 0;
 		dmd.segment.size = dmdi.segment.size;
 
@@ -1859,6 +2176,153 @@ int LUKS2_activate(struct crypt_device *cd,
 	return r;
 }
 
+static bool is_reencryption_helper(const char *name)
+{
+	size_t len;
+
+	if (!name)
+		return false;
+
+	len = strlen(name);
+	return (len >= 9 && (!strncmp(name + len - 8, "-hotzone-", 9) ||
+			     !strcmp(name + len - 8, "-overlay")));
+
+}
+
+static bool contains_reencryption_helper(char **names)
+{
+	while (*names) {
+		if (is_reencryption_helper(*names++))
+			return true;
+	}
+
+	return false;
+}
+
+int LUKS2_deactivate(struct crypt_device *cd, const char *name, struct luks2_hdr *hdr, struct crypt_dm_active_device *dmd, uint32_t flags)
+{
+	int r, ret;
+	struct dm_target *tgt;
+	crypt_status_info ci;
+	struct crypt_dm_active_device dmdc;
+	char **dep, uuid[37], deps_uuid_prefix[40], *deps[MAX_DM_DEPS+1] = { 0 };
+	const char *namei = NULL;
+	struct crypt_lock_handle *reencrypt_lock = NULL;
+
+	if (!dmd || !dmd->uuid)
+		return -EINVAL;
+
+	r = snprintf(deps_uuid_prefix, sizeof(deps_uuid_prefix), CRYPT_SUBDEV "-%.32s", dmd->uuid + 6);
+	if (r < 0 || (size_t)r != (sizeof(deps_uuid_prefix) - 1))
+		return -EINVAL;
+
+	r = snprintf(uuid, sizeof(uuid), "%.8s-%.4s-%.4s-%.4s-%.12s",
+		 dmd->uuid + 6, dmd->uuid + 14, dmd->uuid + 18, dmd->uuid + 22, dmd->uuid + 26);
+	if (r < 0 || (size_t)r != (sizeof(uuid) - 1))
+		return -EINVAL;
+
+	/* uuid mismatch with metadata (if available) */
+	if (hdr && strcmp(hdr->uuid, uuid))
+		return -EINVAL;
+
+	tgt = &dmd->segment;
+
+	/* TODO: We have LUKS2 dependencies now */
+	if (hdr && single_segment(dmd) && tgt->type == DM_CRYPT && crypt_get_integrity_tag_size(cd))
+		namei = device_dm_name(tgt->data_device);
+
+	r = dm_device_deps(cd, name, deps_uuid_prefix, deps, ARRAY_SIZE(deps));
+	if (r < 0)
+		goto out;
+
+	if (contains_reencryption_helper(deps)) {
+		r = crypt_reencrypt_lock(cd, uuid, &reencrypt_lock);
+		if (r) {
+			if (r == -EBUSY)
+				log_err(cd, _("Reencryption in-progress. Cannot deactivate device."));
+			else
+				log_err(cd, _("Failed to get reencryption lock."));
+			goto out;
+		}
+	}
+
+	dep = deps;
+	while (*dep) {
+		if (is_reencryption_helper(*dep) && (dm_status_suspended(cd, *dep) > 0)) {
+			if (dm_error_device(cd, *dep))
+				log_err(cd, _("Failed to replace suspended device %s with dm-error target."), *dep);
+		}
+		dep++;
+	}
+
+	r = dm_query_device(cd, name, DM_ACTIVE_CRYPT_KEY | DM_ACTIVE_CRYPT_KEYSIZE, &dmdc);
+	if (r < 0) {
+		memset(&dmdc, 0, sizeof(dmdc));
+		dmdc.segment.type = DM_UNKNOWN;
+	}
+
+	/* Remove top level device first */
+	r = dm_remove_device(cd, name, flags);
+	if (!r) {
+		tgt = &dmdc.segment;
+		while (tgt) {
+			if (tgt->type == DM_CRYPT)
+				crypt_drop_keyring_key_by_description(cd, tgt->u.crypt.vk->key_description, LOGON_KEY);
+			tgt = tgt->next;
+		}
+	}
+	dm_targets_free(cd, &dmdc);
+
+	/* TODO: We have LUKS2 dependencies now */
+	if (r >= 0 && namei) {
+		log_dbg(cd, "Deactivating integrity device %s.", namei);
+		r = dm_remove_device(cd, namei, 0);
+	}
+
+	if (!r) {
+		ret = 0;
+		dep = deps;
+		while (*dep) {
+			log_dbg(cd, "Deactivating LUKS2 dependent device %s.", *dep);
+			r = dm_query_device(cd, *dep, DM_ACTIVE_CRYPT_KEY | DM_ACTIVE_CRYPT_KEYSIZE, &dmdc);
+			if (r < 0) {
+				memset(&dmdc, 0, sizeof(dmdc));
+				dmdc.segment.type = DM_UNKNOWN;
+			}
+
+			r = dm_remove_device(cd, *dep, flags);
+			if (r < 0) {
+				ci = crypt_status(cd, *dep);
+				if (ci == CRYPT_BUSY)
+					log_err(cd, _("Device %s is still in use."), *dep);
+				if (ci == CRYPT_INACTIVE)
+					r = 0;
+			}
+			if (!r) {
+				tgt = &dmdc.segment;
+				while (tgt) {
+					if (tgt->type == DM_CRYPT)
+						crypt_drop_keyring_key_by_description(cd, tgt->u.crypt.vk->key_description, LOGON_KEY);
+					tgt = tgt->next;
+				}
+			}
+			dm_targets_free(cd, &dmdc);
+			if (r && !ret)
+				ret = r;
+			dep++;
+		}
+		r = ret;
+	}
+
+out:
+	crypt_reencrypt_unlock(cd, reencrypt_lock);
+	dep = deps;
+	while (*dep)
+		free(*dep++);
+
+	return r;
+}
+
 int LUKS2_unmet_requirements(struct crypt_device *cd, struct luks2_hdr *hdr, uint32_t reqs_mask, int quiet)
 {
 	uint32_t reqs;
@@ -1882,6 +2346,8 @@ int LUKS2_unmet_requirements(struct crypt_device *cd, struct luks2_hdr *hdr, uin
 
 	if (reqs_reencrypt(reqs) && !quiet)
 		log_err(cd, _("Offline reencryption in progress. Aborting."));
+	if (reqs_reencrypt_online(reqs) && !quiet)
+		log_err(cd, _("Online reencryption in progress. Aborting."));
 
 	/* any remaining unmasked requirement fails the check */
 	return reqs ? -EINVAL : 0;
@@ -1933,3 +2399,17 @@ int json_object_object_add_by_uint(json_object *jobj, unsigned key, json_object
 	return 0;
 #endif
 }
+
+/* jobj_dst must contain pointer initialised to NULL (see json-c json_object_deep_copy API) */
+int json_object_copy(json_object *jobj_src, json_object **jobj_dst)
+{
+	if (!jobj_src || !jobj_dst || *jobj_dst)
+		return -1;
+
+#if HAVE_DECL_JSON_OBJECT_DEEP_COPY
+	return json_object_deep_copy(jobj_src, jobj_dst, NULL);
+#else
+	*jobj_dst = json_tokener_parse(json_object_get_string(jobj_src));
+	return *jobj_dst ? 0 : -1;
+#endif
+}
