@@ -24,17 +24,32 @@
 
 #include <stddef.h>
 
+typedef enum { LOGON_KEY = 0, USER_KEY } key_type_t;
+
+const char *key_type_name(key_type_t ktype);
+
 int keyring_check(void);
 
+int keyring_get_key(const char *key_desc,
+		    char **key,
+		    size_t *key_size);
+
 int keyring_get_passphrase(const char *key_desc,
 		      char **passphrase,
 		      size_t *passphrase_len);
 
 int keyring_add_key_in_thread_keyring(
+	key_type_t ktype,
 	const char *key_desc,
 	const void *key,
 	size_t key_size);
 
-int keyring_revoke_and_unlink_key(const char *key_desc);
+int keyring_add_key_in_user_keyring(
+	key_type_t ktype,
+	const char *key_desc,
+	const void *key,
+	size_t key_size);
+
+int keyring_revoke_and_unlink_key(key_type_t ktype, const char *key_desc);
 
 #endif
